import { Injectable, Logger } from '@nestjs/common';
import { InjectModel } from '@nestjs/mongoose';
import { Model } from 'mongoose';
import { EspaceCoopCNFS, EspaceCoopCNFSDocument } from '../schemas/espaceCoopCNFS.schema';
import { lastValueFrom } from 'rxjs';
import { Cron, CronExpression } from '@nestjs/schedule';
import { HttpService } from '@nestjs/axios';
import { UsersService } from '../../users/services/users.service';
import { Permanence } from '../interfaces/permanence.interface';

@Injectable()
export class EspaceCoopService {
  private readonly logger = new Logger(EspaceCoopService.name);

  constructor(
    @InjectModel(EspaceCoopCNFS.name) private espaceCoopCNFSModel: Model<EspaceCoopCNFSDocument>,
    private httpService: HttpService,
    private readonly userService: UsersService
  ) {}

  public async findAll(): Promise<EspaceCoopCNFS[]> {
    this.logger.debug('Getting all CNFS from espace coop stored in DB...');
    return this.espaceCoopCNFSModel.find().exec();
  }

  @Cron(CronExpression.EVERY_WEEK)
  public async loadEspaceCoopCNFS(): Promise<EspaceCoopCNFS[]> {
    this.logger.debug('Fetching all CNFS from espace coop...');

    try {
      const allPermanences = await lastValueFrom(
        this.httpService.get<Permanence[]>('https://api.conseiller-numerique.gouv.fr/permanences')
      );
      const rhonePermanences = allPermanences.data.filter((permanence) => permanence.code_postal.startsWith('69'));
      this.logger.debug('Parsing CNFS from espace coop...');
      const espaceCoopCNFS = await this.parsePermanences(rhonePermanences);

      this.logger.debug('Clearing previous CNFS...');
      await this.espaceCoopCNFSModel.deleteMany({}).exec();

      this.logger.debug('Saving new CNFS...');
      return this.espaceCoopCNFSModel.create(espaceCoopCNFS);
    } catch (error) {
      this.logger.error('Error:', error);
      throw new Error('Failed to load and save EspaceCoopCNFS data.');
    }
  }

  public async parsePermanences(permanences: Permanence[]): Promise<EspaceCoopCNFS[]> {
    const cnfsMap = new Map<string, EspaceCoopCNFS>();

    for (const permanence of permanences) {
      for (const aidant of permanence.aidants) {
        const { nom, courriel, telephone } = aidant;
        if (cnfsMap.has(nom)) {
          const cnfs = cnfsMap.get(nom);
          if (!cnfs.structures.includes(permanence.nom)) {
            cnfs.structures.push(permanence.nom);
          }
        } else {
          const resinUser = await this.userService.findOne(courriel);
          cnfsMap.set(nom, {
            name: nom,
            email: courriel ?? null,
            phone: telephone ?? null,
            employer: permanence.structureNom,
            structures: [permanence.nom],
            hasAccount: Boolean(resinUser),
          });
        }
      }
    }

    return Array.from(cnfsMap.values());
  }
}
