import { Test, TestingModule } from '@nestjs/testing';
import { getModelToken } from '@nestjs/mongoose';
import { StructuresExportService } from './structures-export.service';
import { Model, Query } from 'mongoose';
import { StructureDocument } from '../schemas/structure.schema';
import { StructureFormatted } from '../interfaces/structure-formatted.interface';
import { mockFormattedStructures } from '../../../test/mock/services/structures-export.mock.service';
import { StructureCategories } from '../schemas/structureCategories.schema';
import { StructuresService } from './structures.service';

const mockStructureModel = {
  create: jest.fn(),
  deleteOne: jest.fn(),
  countDocuments: jest.fn(),
  findOne: jest.fn(),
  findById: jest.fn(),
  limit: jest.fn(),
  findByIdAndUpdate: jest.fn(),
  setCategoriesWithPersonalOffers: jest.fn(),
  exec: jest.fn(),
  find: jest.fn(),
  save: jest.fn(),
};

function createMockedStructure(overrides: Partial<StructureDocument> = {}) {
  const defaultStructure = {
    id: '6093ba0e2ab5775cfc01abcd',
    coord: [4.8608585, 45.7086482],
    categories: {
      accessModality: ['accesLibre'],
      labelsQualifications: ['pix'],
      onlineProcedures: [],
      selfServiceMaterial: ['computer', 'scanner'],
      solidarityMaterial: ['computer'],
      baseSkills: ['computer', 'internet', 'smartphone', 'mail', 'work', 'officeAutomation', 'communication'],
      advancedSkills: [
        'cyberAndConf',
        'parenting',
        'videoGames',
        'numericCreate',
        'fakeNews',
        'photoVideo',
        'devAndRobots',
      ],
      age: ['young', 'old'],
      languageAndIlliteracy: [],
      handicaps: ['physicalDisability'],
      genre: ['uniquementFemmes'],
    },
    structureType: null,
    structureName: 'A. Le Son du Clic',
    description:
      "Entreprise de médiation numérique assurant l'animation d'atelier numériques en collectif ou l'accompagnement au numérique des particuliers en individuel. Structure agréée APTIC sur la métropole de Lyon, acceptant à ce titre le règlement de ses prestations en chèque PASS NUMERIQUE",
    lockdownActivity: null,
    address: {
      numero: '7',
      street: 'Rue Robert et Reynier ',
      commune: 'Saint-Fons',
      inseeCode: '69204',
      postcode: '69190',
    },
    contactPhone: '06 60 60 60 60',
    contactMail: 'adzadadzadazd@pimms.org',
    website: 'http://lesonduclic.fr/site1/',
    facebook: null,
    twitter: null,
    instagram: null,
    linkedin: null,
    hours: {
      monday: {
        open: true,
        time: [
          {
            closing: '17:00',
            opening: '09:00',
          },
        ],
      },
      tuesday: {
        open: true,
        time: [
          {
            closing: '17:00',
            opening: '09:00',
          },
        ],
      },
      wednesday: {
        open: true,
        time: [
          {
            closing: '17:00',
            opening: '09:00',
          },
        ],
      },
      thursday: {
        open: true,
        time: [
          {
            closing: '17:00',
            opening: '09:00',
          },
        ],
      },
      friday: {
        open: true,
        time: [
          {
            closing: '17:00',
            opening: '09:00',
          },
        ],
      },
      saturday: {
        open: false,
        time: [],
      },
      sunday: {
        open: false,
        time: [],
      },
    },
    pmrAccess: false,
    exceptionalClosures: null,
    otherDescription: null,
    nbComputers: 1,
    nbPrinters: 1,
    nbScanners: 1,
    freeWorkShop: 'Oui',
    accountVerified: true,
    personalOffers: [],
    dataShareConsentDate: new Date('2021-01-24T09:59:26.000Z'),
    createdAt: new Date('2021-05-06T09:42:38.000Z'),
    updatedAt: new Date('2023-03-01T09:42:50.000Z'),
    __v: 0,
  };

  return { ...defaultStructure, ...overrides };
}

describe('StructuresExportService', () => {
  let service: StructuresExportService;
  let structureModel: Model<StructureDocument>;

  beforeEach(async () => {
    const module: TestingModule = await Test.createTestingModule({
      providers: [
        StructuresExportService,
        {
          provide: getModelToken('Structure'),
          useValue: {
            find: jest.fn(),
          },
        },
        {
          provide: StructuresService,
          useValue: mockStructureModel,
        },
      ],
    }).compile();

    service = module.get<StructuresExportService>(StructuresExportService);
    structureModel = module.get<Model<StructureDocument>>(getModelToken('Structure'));
  });

  describe('exportFormatted', () => {
    it('should return formatted structures', async () => {
      const rawStructure = createMockedStructure();
      rawStructure['categoriesWithPersonalOffers'] = rawStructure.categories;
      const mockQuery = ({
        select: jest.fn().mockReturnThis(),
        exec: jest.fn().mockResolvedValueOnce([rawStructure]),
        populate: jest.fn().mockReturnThis(),
      } as unknown) as Query<StructureDocument[], StructureDocument>;

      jest.spyOn(structureModel, 'find').mockReturnValueOnce(mockQuery);

      const expectedFormattedStructure: StructureFormatted = mockFormattedStructures[0];
      const result: StructureFormatted[] = await service.exportFormatted();

      expect(structureModel.find).toHaveBeenCalledWith({
        deletedAt: { $exists: false },
        accountVerified: true,
        $where: 'this.dataShareConsentDate != null',
      });
      expect(result).toEqual([expectedFormattedStructure]);
    });
  });

  describe('checkPublic', () => {
    const invokeCheckPublic = (age: string[], handicaps: string[], languages: string[], other: string[]) =>
      (service as any).checkPublic(age, handicaps, languages, other); // any to bypass ts access checks on private method
    it('should return an empty string when no parameters are provided', () => {
      const result = invokeCheckPublic([], [], [], []);
      expect(result).toEqual('');
    });

    it('should correctly map age, handicaps, languages, and other parameters to a formatted string', () => {
      const age = ['family', 'young'];
      const handicaps = ['blindness', 'mentalDisability'];
      const languages = ['illettrisme', 'english'];
      const other = ['uniquementFemmes'];

      const result = invokeCheckPublic(age, handicaps, languages, other);

      expect(result).toEqual(
        'Familles/enfants;Jeunes (16-26 ans);Déficience visuelle;Handicaps mentaux : déficiences limitant les activités d’une personne;Personnes en situation d’illettrisme;Public langues étrangères;Uniquement femmes'
      );
    });

    it('should ignore undefined values and return a formatted string without them', () => {
      const age = ['family', 'young'];
      const handicaps = ['blindness', undefined, 'mentalDisability'];
      const languages = ['illettrisme', undefined, 'english'];
      const other = ['uniquementFemmes', undefined];

      const result = invokeCheckPublic(age, handicaps, languages, other);

      expect(result).toEqual(
        'Familles/enfants;Jeunes (16-26 ans);Déficience visuelle;Handicaps mentaux : déficiences limitant les activités d’une personne;Personnes en situation d’illettrisme;Public langues étrangères;Uniquement femmes'
      );
    });
  });

  describe('checkAccessConditions', () => {
    const invokeCheckAccessConditions = (accessModality: string[]) =>
      (service as any).checkAccessConditions(accessModality); // any to bypass ts access checks on private method

    it('should return an empty string when no access modality is provided', () => {
      const result = invokeCheckAccessConditions([]);
      expect(result).toEqual('');
    });

    it('should return the correct access condition for "accesLibre"', () => {
      const accessModality = ['accesLibre'];
      const result = invokeCheckAccessConditions(accessModality);
      expect(result).toEqual('Gratuit : Je peux accéder gratuitement au lieu et à ses services');
    });

    it('should return an empty string for unknown access modalities', () => {
      const accessModality = ['unknown'];
      const result = invokeCheckAccessConditions(accessModality);
      expect(result).toEqual('');
    });
  });

  describe('formatServices', () => {
    const invokeFormatServices = (structCat: StructureCategories) => (service as any).formatServices(structCat); // any to bypass ts access checks on private method

    it('should return an empty string when no structure categories are provided', () => {
      const result = invokeFormatServices(null);
      expect(result).toEqual('');
    });

    it('should return the formatted services when matching categories and values are found', () => {
      const structCat = {
        selfServiceMaterial: ['computer', 'scanner'],
        solidarityMaterial: ['computer'],
        baseSkills: ['computer', 'internet', 'smartphone', 'mail', 'work', 'officeAutomation', 'communication'],
        advancedSkills: [
          'cyberAndConf',
          'parenting',
          'videoGames',
          'numericCreate',
          'fakeNews',
          'photoVideo',
          'devAndRobots',
        ],
      };
      const result = invokeFormatServices(structCat);
      expect(result).toEqual(
        "Prendre en main un smartphone ou une tablette;Prendre en main un ordinateur;Utiliser le numérique au quotidien;Approfondir ma culture numérique;Favoriser mon insertion professionnelle;Accéder à du matériel;S’équiper en matériel informatique;Créer avec le numérique;Soutenir la parentalité et l'éducation avec le numérique"
      );
    });

    it('should return an empty string when no matching categories and values are found', () => {
      const structCat = {
        category1: ['value2'],
        category3: ['value4'],
      };
      const result = invokeFormatServices(structCat);
      expect(result).toEqual('');
    });
  });

  describe('formatAddress', () => {
    const invokeFormatAddress = (address: { numero?: string; street: string }) =>
      (service as any).formatAddress(address); // any to bypass ts access checks on private method

    it('should return the formatted address with street only when no numero is provided', () => {
      const address = { street: 'Rue Robert et Reynier' };
      const result = invokeFormatAddress(address);
      expect(result).toEqual('Rue Robert et Reynier');
    });

    it('should return the formatted address with numero and street when numero is provided', () => {
      const address = { numero: '123', street: 'Rue Robert et Reynier' };
      const result = invokeFormatAddress(address);
      expect(result).toEqual('123 Rue Robert et Reynier');
    });
  });

  describe('convertToOSMFormat', () => {
    const invokeConvertToOSMFormat = (openingHours: Record<string, any>) =>
      (service as any).convertToOSMFormat(openingHours); // any to bypass ts access checks on private method

    it('should return empty string when openingHours is empty', () => {
      const openingHours = {
        monday: {
          open: false,
          time: [],
        },
        tuesday: {
          open: false,
          time: [],
        },
        wednesday: {
          open: false,
          time: [],
        },
        thursday: {
          open: false,
          time: [],
        },
        friday: {
          open: false,
          time: [],
        },
        saturday: {
          open: false,
          time: [],
        },
        sunday: {
          open: false,
          time: [],
        },
      };
      const result = invokeConvertToOSMFormat(openingHours);
      expect(result).toEqual('');
    });

    it('should return the converted OSM opening hours format for the provided openingHours', () => {
      const openingHours = {
        monday: { open: true, time: [{ opening: '09:00', closing: '17:00' }] },
        tuesday: { open: true, time: [{ opening: '09:00', closing: '17:00' }] },
        wednesday: { open: true, time: [{ opening: '09:00', closing: '17:00' }] },
        thursday: { open: true, time: [{ opening: '09:00', closing: '17:00' }] },
        friday: { open: true, time: [{ opening: '09:00', closing: '17:00' }] },
        saturday: { open: false },
        sunday: { open: false },
      };
      const result = invokeConvertToOSMFormat(openingHours);
      expect(result).toEqual('Mo-Fr 09:00-17:00');
    });
  });
});
