import { Injectable, Logger } from '@nestjs/common';
import { StructureFormatted } from '../interfaces/structure-formatted.interface';
import { HttpService } from '@nestjs/axios';
import { Model } from 'mongoose';
import { Structure, StructureDocument } from '../schemas/structure.schema';
import { InjectModel } from '@nestjs/mongoose';
import { Cron, CronExpression } from '@nestjs/schedule';
import { lastValueFrom } from 'rxjs';
import * as config from './structures-export-configs/services-config.json';
import { Week } from '../../shared/schemas/week.schema';
import { Day } from '../schemas/day.schema';
import { Time } from '../schemas/time.schema';
import { DataGouvStructure } from '../interfaces/data-gouv-structure.interface';

interface Condition {
  address?: { street?: string; numero?: string };
  contactMail?: string;
}

@Injectable()
export class StructuresImportService {
  private readonly logger = new Logger(StructuresImportService.name);
  constructor(
    private readonly httpService: HttpService,
    @InjectModel(Structure.name) private structureModel: Model<StructureDocument>
  ) {}

  @Cron(CronExpression.EVERY_DAY_AT_MIDNIGHT)
  public async importDataGouvStructures(): Promise<void> {
    this.logger.debug(`Importing all structures from data.gouv dataset...`);
    try {
      const dataset = await this.fetchDataset();
      const rhoneStructures = this.filterStructuresByCode(dataset, '69');
      this.logger.debug(`Found ${rhoneStructures.length} structures from rhône in data.gouv dataset...`);
      await this.processStructures(rhoneStructures);
    } catch (error) {
      this.logger.error(`Error while importing data.gouv dataset: ${error}`);
    }
  }

  private async fetchDataset(): Promise<DataGouvStructure[]> {
    const response = await lastValueFrom(
      this.httpService.get<DataGouvStructure[]>(
        'https://www.data.gouv.fr/fr/datasets/r/bdfef095-f3a9-4e66-84bb-ea05d5c9c3a5'
      )
    );
    return response.data;
  }

  private filterStructuresByCode(structures: DataGouvStructure[], code: string): DataGouvStructure[] {
    return structures.filter((struct) => struct.code_postal.startsWith(code));
  }

  private async processStructures(structures: DataGouvStructure[]): Promise<void> {
    let newOne = 0;
    let skipped = 0;
    for (const structure of structures) {
      if (await this.doesAlreadyExist(structure)) {
        skipped++;
      } else {
        await this.createStructure(structure);
        newOne++;
      }
    }
    this.logger.debug(`Created ${newOne} new structures and skipped ${skipped} structures...`);
  }

  private async createStructure(structure: DataGouvStructure): Promise<void> {
    const formattedStructure = await this.formatToResinSchema(structure);
    const newStructure = new this.structureModel(formattedStructure);
    await newStructure.save();
  }

  private async doesAlreadyExist(structure: DataGouvStructure): Promise<boolean> {
    if (structure.source === 'res-in') {
      return true;
    }

    let addressCondition = {};
    // test if string starts with a number (22, 13bis ...)
    if (/^\d/.test(structure.adresse)) {
      const addressParts = structure.adresse.split(' ');
      const numero = addressParts.shift();
      const street = addressParts.join(' ');
      addressCondition = { 'address.street': street, 'address.numero': numero };
    } else {
      addressCondition = { 'address.street': structure.adresse };
    }

    const conditions: Condition[] = [addressCondition, { contactMail: structure.courriel }];
    const count = await this.structureModel.countDocuments({ $and: conditions }).exec();
    return count > 0;
  }

  private async formatToResinSchema(structure: DataGouvStructure): Promise<Structure> {
    const formatted = new Structure({
      structureName: structure.nom,
      address: this.formatAddress(structure.adresse, structure.commune, structure.code_postal, structure.code_insee),
      updatedAt: new Date(structure.date_maj),
      categories: this.servicesToCategories(
        structure.services,
        structure.conditions_acces,
        structure.publics_accueillis
      ),
      coord: [structure.longitude, structure.latitude],
      description: structure.presentation_resume ?? null,
      contactPhone: structure.telephone ?? null,
      contactMail: structure.courriel ?? null,
      hours: this.convertFromOSMFormat(structure.horaires),
      accountVerified: true,
      remoteAccompaniment: false,
      pmrAccess: false,
      nbComputers: 0,
      nbPrinters: 0,
      nbScanners: 0,
      freeWorkShop: 'Non',
      lockdownActivity: null,
      structureType: null,
      website: structure.site_web ?? null,
      dataShareConsentDate: null,
      exceptionalClosures: null,
      facebook: null,
      instagram: null,
      numero: '',
      twitter: null,
      personalOffers: [],
      otherDescription: null,
      linkedin: null,
    });

    return formatted;
  }

  private formatAddress(
    adresse: StructureFormatted['adresse'],
    commune: StructureFormatted['commune'],
    code_postal: StructureFormatted['code_postal'],
    code_insee: StructureFormatted['code_insee']
  ): StructureDocument['address'] {
    let numero = null;
    let street = null;

    // test if string starts with a number (22, 13bis ...)
    if (/^\d/.test(adresse)) {
      const addressParts = adresse.split(' ');
      numero = addressParts.shift();
      street = addressParts.join(' ');
    } else {
      street = adresse;
    }
    return {
      numero: numero,
      street: street,
      commune: commune,
      postcode: code_postal,
      inseeCode: code_insee,
    };
  }

  private servicesToCategories(
    services: StructureFormatted['services'],
    conditions_acces: StructureFormatted['conditions_acces'],
    publics_accueillis: StructureFormatted['publics_accueillis']
  ): StructureDocument['categories'] {
    const categories: StructureDocument['categories'] = {
      labelsQualifications: [],
      accessModality: [],
      onlineProcedures: [],
      selfServiceMaterial: [],
      equipmentsServices: [],
      baseSkills: [],
      advancedSkills: [],
      age: [],
      languageAndIlliteracy: [],
      handicaps: [],
      solidarityMaterial: [],
      ctm: [],
      genre: [],
    };
    if (!services) {
      return categories;
    }

    const servicesArray = services.split(';');

    for (const [stringToMatch, serviceConfig] of Object.entries(config)) {
      if (servicesArray.includes(stringToMatch)) {
        for (const category of serviceConfig.categories) {
          categories[category] = [...new Set([...categories[category], ...serviceConfig.values])];
        }
      }
    }

    if (conditions_acces === 'Gratuit : Je peux accéder gratuitement au lieu et à ses services') {
      categories.accessModality.push('accesLibre');
    }

    const ageMappingReverse = {
      'Familles/enfants': 'family',
      'Jeunes (16-26 ans)': 'young',
      Adultes: 'adults',
      'Seniors (+ 65 ans)': 'old',
    };

    const handicapsMappingReverse = {
      'Déficience visuelle': 'blindness',
      Surdité: 'deafness',
      'Handicaps mentaux : déficiences limitant les activités d’une personne': 'mentalDisability',
    };

    const languageMappingReverse = {
      'Personnes en situation d’illettrisme': 'illettrisme',
      'Public langues étrangères': 'foreignLanguage',
    };

    const otherMappingReverse = {
      'Uniquement femmes': 'uniquementFemmes',
    };

    if (publics_accueillis !== undefined) {
      const allPublics = publics_accueillis.split(';');
      allPublics.forEach((value) => {
        if (value in ageMappingReverse) {
          categories.age.push(ageMappingReverse[value]);
        } else if (value in handicapsMappingReverse) {
          categories.handicaps.push(handicapsMappingReverse[value]);
        } else if (value in languageMappingReverse) {
          categories.languageAndIlliteracy.push(languageMappingReverse[value]);
        } else if (value in otherMappingReverse) {
          categories.genre.push(otherMappingReverse[value]);
        }
      });
    }
    return categories;
  }

  private convertFromOSMFormat(osmHours: string): Week {
    const daysOfWeek = ['mo', 'tu', 'we', 'th', 'fr', 'sa', 'su'];
    const fullDaysOfWeek = ['monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday', 'sunday'];
    const week = new Week();

    // initialize all days as closed
    fullDaysOfWeek.forEach((day) => {
      week[day] = new Day();
      week[day].open = false;
      week[day].time = [];
    });

    if (osmHours === undefined) {
      return week;
    }

    const entries = osmHours.split(';');

    entries.forEach((entry) => {
      // split by ',' to get individual time ranges for each day(s)
      const timeRanges = entry.split(',');

      timeRanges.forEach((timeRange) => {
        // split by ' ' to separate days and time
        const [days, times] = timeRange.trim().split(' ');

        // check if days contains ',' then handle as individual days
        if (days.includes(',')) {
          const individualDays = days.split(',');
          individualDays.forEach((day) => {
            const startIndex = daysOfWeek.indexOf(day.toLowerCase());
            if (times) {
              const [opening, closing] = times.split('-');
              week[fullDaysOfWeek[startIndex]].open = true;
              week[fullDaysOfWeek[startIndex]].time.push({ opening, closing } as Time);
            }
          });
        } else {
          // split days by '-' to handle ranges (Tu-We)
          const [startDay, endDay] = days.toLowerCase().split('-');

          // find start and end indexes in the week for the range
          const startIndex = daysOfWeek.indexOf(startDay);
          let endIndex = daysOfWeek.indexOf(endDay);

          // if endDay is not provided, assume it's the same as startDay
          if (endIndex === -1) {
            endIndex = startIndex;
          }

          // iterate over the range of days and set the hours
          if (times) {
            const [opening, closing] = times.split('-');
            for (let i = startIndex; i <= endIndex; i++) {
              week[fullDaysOfWeek[i]].open = true;
              week[fullDaysOfWeek[i]].time.push({ opening, closing } as Time);
            }
          }
        }
      });
    });

    return week;
  }
}
