import { Logger } from '@nestjs/common';
import { ConfigModule } from '@nestjs/config';
import { Test, TestingModule } from '@nestjs/testing';
import { StructuresForSearchServiceMock } from '../../../test/mock/services/structures-for-search.mock.service';
import { MailerModule } from '../../mailer/mailer.module';
import { SearchModule } from '../../search/search.module';
import { StructuresSearchService } from './structures-search.service';
import { StructuresService } from './structures.service';
describe('StructuresSearchService', () => {
  let service: StructuresSearchService;

  beforeEach(async () => {
    const module: TestingModule = await Test.createTestingModule({
      imports: [MailerModule, SearchModule, ConfigModule],
      providers: [
        StructuresSearchService,
        {
          provide: StructuresService,
          useClass: StructuresForSearchServiceMock,
        },
      ],
    }).compile();

    service = module.get<StructuresSearchService>(StructuresSearchService);
    service['index'] = 'structures-unit-test';
    await service.dropIndex();
    await service.createStructureIndex();

    // init str
    const structuresForSearchService = new StructuresForSearchServiceMock();
    const structures = structuresForSearchService.findAll();
    await Promise.all(
      structures.map((structure) => {
        service.indexStructure(structure);
      })
    );

    // wait for the new structures to be indexed before search
    await service.refreshIndexStructure();
    await new Promise((r) => setTimeout(r, 2000));
  }, 10000);

  it('should be defined', () => {
    expect(service).toBeDefined();
  });

  it('should find maisons de la métropole', async () => {
    const res = await service.search('maison de la');
    //Logger.log(JSON.stringify(res));
    expect(res[0].structureName).toContain('Maison de la Métropole');
    expect(res[1].structureName).toContain('Maison de la Métropole');
  });

  it('should find metropole', async () => {
    const res = await service.search('metropole');
    expect(res[0].structureName).toContain('Métropole');
  });

  it('should find structure with stopword "son"', async () => {
    const res = await service.search('son');
    expect(res[0].structureName).toContain('Le Son du Clic');
  });

  it('should find structure with stopword "mais"', async () => {
    const res = await service.search('mais');
    expect(res[0].structureName.toLowerCase()).toContain('maison');
  });

  it('should find structure with stopword "le"', async () => {
    const res = await service.search('le');
    expect(res.length).toBeGreaterThanOrEqual(1);
  });

  it('should find structure with partial word "oull"', async () => {
    const res = await service.search('oull');
    expect(res[0].structureName.toLowerCase()).toContain('oullins');
  });

  it('should find structure with partial word "vill"', async () => {
    const res = await service.search('vill');
    expect(res[0].structureName.toLowerCase()).toContain('villeurbanne');
  });

  it('should find structure with partial word "asso"', async () => {
    const res = await service.search('asso');
    Logger.log(JSON.stringify(res));
    expect(res[0].structureName.toLowerCase()).toContain('association');
  });

  it('should find text in description', async () => {
    const res = await service.search('liseuse');
    expect(res.length).toBe(1);
    expect(res[0].structureName).toContain("Médiathèque d'Ecully");
  });

  it('should not find text in description when search by name', async () => {
    const res = await service.search('liseuse', ['structureName']);
    expect(res.length).toBe(0);
  });

  it('should find structure when search by name', async () => {
    const res = await service.search('le son', ['structureName']);
    expect(res[0].structureName).toContain('Le Son du Clic');
  });

  it('should drop index', async () => {
    const res = await service.dropIndex();
    expect(res).toBeTruthy();
  });
});
