import { HttpService } from '@nestjs/axios';
import { HttpStatus } from '@nestjs/common';
import { Test, TestingModule } from '@nestjs/testing';
import { Types } from 'mongoose';
import { CategoriesServiceMock } from '../../test/mock/services/categories.mock.service';
import { HttpServiceMock } from '../../test/mock/services/http.mock.service';
import { StructuresServiceMock } from '../../test/mock/services/structures.mock.service';
import { TempUserServiceMock } from '../../test/mock/services/tempUser.mock.service';
import { UsersServiceMock } from '../../test/mock/services/user.mock.service';
import { CategoriesService } from '../categories/services/categories.service';
import { PersonalOffersService } from '../personal-offers/personal-offers.service';
import { TempUserService } from '../temp-user/temp-user.service';
import { UserRole } from '../users/enum/user-role.enum';
import { UsersService } from '../users/services/users.service';
import { PersonalOffersServiceMock } from './../../test/mock/services/personalOffers.mock.service';
import { PersonalOffer, PersonalOfferDocument } from './../personal-offers/schemas/personal-offer.schema';
import { CreateStructureDto } from './dto/create-structure.dto';
import { StructuresService } from './services/structures.service';
import { StructuresController } from './structures.controller';
import { StructuresExportService } from './services/structures-export.service';
import {
  StructuresExportServiceMock,
  mockFormattedStructures,
} from '../../test/mock/services/structures-export.mock.service';
import { StructureFormatted } from './interfaces/structure-formatted.interface';
import { JobDocument } from '../users/schemas/job.schema';
describe('StructuresController', () => {
  let controller: StructuresController;
  let usersService: UsersService;
  let structuresExportService: StructuresExportService;
  let personalOffersService: PersonalOffersService;

  beforeEach(async () => {
    const module: TestingModule = await Test.createTestingModule({
      imports: [],
      controllers: [StructuresController],
      providers: [
        {
          provide: HttpService,
          useClass: HttpServiceMock,
        },
        {
          provide: StructuresService,
          useClass: StructuresServiceMock,
        },
        {
          provide: UsersService,
          useClass: UsersServiceMock,
        },
        {
          provide: PersonalOffersService,
          useClass: PersonalOffersServiceMock,
        },
        {
          provide: TempUserService,
          useClass: TempUserServiceMock,
        },
        {
          provide: CategoriesService,
          useClass: CategoriesServiceMock,
        },
        {
          provide: StructuresExportService,
          useClass: StructuresExportServiceMock,
        },
      ],
    }).compile();
    controller = module.get<StructuresController>(StructuresController);
    usersService = module.get<UsersService>(UsersService);
    structuresExportService = module.get<StructuresExportService>(StructuresExportService);
    personalOffersService = module.get<PersonalOffersService>(PersonalOffersService);
  });

  it('should be defined', () => {
    expect(controller).toBeDefined();
  });

  it('should get structure coordinates', async () => {
    const coords = controller.getCoordinates('Lyon');
    expect(coords).toBeTruthy();
  });

  it('should create structure', async () => {
    const structure: CreateStructureDto = {
      userEmail: 'admin@admin.com',
      structure: null,
    };
    const res = controller.create(structure);
    expect(res).toBeTruthy();
  });

  it('should update structure after ownerVerify', async () => {
    const structureId = '1';
    const res = controller.updateAfterOwnerVerify(structureId);
    expect(res).toBeTruthy();
  });

  it('should update structure', async () => {
    const structureService = new StructuresServiceMock();
    const structure = structureService.findOne(new Types.ObjectId('6093ba0e2ab5775cfc01ed3e'));

    const structureId = '1';
    const res = await controller.update(structureId, {
      deletedAt: null,
      remoteAccompaniment: null,
      dataShareConsentDate: null,
      personalOffers: [],
      ...structure,
    });
    expect(res.structureName).toBe('a');
  });

  it('should get all structure', async () => {
    const res = await controller.findAll();
    expect(res.length).toBe(2);
  });

  it('should get all Formatted structure', async () => {
    const mockResult: StructureFormatted[] = mockFormattedStructures;
    jest.spyOn(structuresExportService, 'exportFormatted').mockResolvedValueOnce(mockResult);

    const result: StructureFormatted[] = await controller.findAllFormatted();

    expect(result).toEqual(mockResult);
  });

  it('should search structure', async () => {
    const res = controller.search(null, null);
    expect(res).toBeTruthy();
  });

  it('should reset Search Index', async () => {
    const res = controller.resetES();
    expect(res).toBeTruthy();
  });

  it('should claim structure', async () => {
    const userMock = new UsersServiceMock();
    const user = userMock.findOne('pauline.dupont@mii.com');
    const res = controller.claim('1', {
      phone: null,
      resetPasswordToken: null,
      changeEmailToken: null,
      newEmail: null,
      pendingStructuresLink: null,
      structuresLink: null,
      structureOutdatedMailSent: null,
      personalOffers: null,
      email: user.email,
      name: user.name,
      surname: user.surname,
      emailVerified: true,
      createdAt: new Date('2022-05-25T09:48:28.824Z'),
      password: user.password,
      validationToken: null,
      role: null,
      employer: {
        name: 'test',
        validated: true,
      },
      job: {
        name: 'test',
        validated: true,
        hasPersonalOffer: false,
        jobsGroup: null,
      } as JobDocument,
      unattachedSince: null,
    });
    expect(res).toBeTruthy();
  });

  it('should search an address', async () => {
    const res = controller.searchAddress({ searchQuery: 'Rue Alphonse Daudet' });
    expect(res).toBeTruthy();
  });

  describe('find(structureId)', () => {
    it('should find structure', async () => {
      let res = controller.find({ user: {} }, '6093ba0e2ab5775cfc01ed3e');
      expect(res).toBeTruthy();
      res = controller.find({ user: {} }, '');
      expect(res).toBeTruthy();
    });

    it('should find deleted structure as an admin', async () => {
      const response = controller.find({ user: { role: UserRole.admin } }, '6093ba0e2ab5775cfc01ed33');
      expect(response).toBeTruthy();
      expect((await response).deletedAt).toBeTruthy();
    });

    it('should not find deleted structure as anonymous', async () => {
      try {
        await controller.find({ user: {} }, '6093ba0e2ab5775cfc01ed33');
        expect(true).toBe(false);
      } catch (error) {
        expect(error.status).toEqual(HttpStatus.NOT_FOUND);
      }
    });
  });

  it('should find struct with owners', async () => {
    const res = controller.findWithOwners('6093ba0e2ab5775cfc01ed3e');
    expect(res).toBeTruthy();
  });

  it('should delete struct', async () => {
    const req = { user: { _id: '6036721022462b001334c4bb' } };
    const res = controller.delete(req, '6093ba0e2ab5775cfc01ed3e');
    expect(res).toBeTruthy();
  });

  it('should cancel struct delete', async () => {
    const req = { user: { _id: '6036721022462b001334c4bb' } };
    const res = controller.cancelDelete(req, '6093ba0e2ab5775cfc01ed3e');
    expect(res).toBeTruthy();
  });

  it('should add Owner', async () => {
    let res = controller.addOwner('6093ba0e2ab5775cfc01ed3e', { email: 'pauline.dupont@mii.com' });
    expect(res).toBeTruthy();
    res = controller.addOwner('6093ba0e2ab5775cfc01ed3e', { email: 'pauline.dupont@mii.fr' });
    expect(res).toBeTruthy();
    res = controller.addOwner('', { email: 'pauline.dupont@mii.fr' });
    expect(res).toBeTruthy();
  });

  describe('removeOwner', () => {
    it('should return invalid structure', async () => {
      try {
        await controller.removeOwner('invalidstructure', '63639058685ba134c32bc495');
        expect(true).toBe(false);
      } catch (e) {
        expect(e.message).toEqual('Invalid Structure');
        expect(e.status).toEqual(HttpStatus.NOT_FOUND);
      }
    });
    it('should return invalid user', async () => {
      try {
        await controller.removeOwner('6093ba0e2ab5775cfc01ed3e', 'invaliduser');
        expect(true).toBe(false);
      } catch (e) {
        expect(e.message).toEqual('Invalid User');
        expect(e.status).toEqual(HttpStatus.NOT_FOUND);
      }
    });
    it('should remove structure link without offer', async () => {
      jest.spyOn(usersService, 'getPersonalOfferInStructure').mockImplementationOnce(() => null);
      const removeFromStructureLinkedSpyer = jest.spyOn(usersService, 'removeFromStructureLinked');
      const deleteOfferSpyer = jest.spyOn(personalOffersService, 'delete');
      await controller.removeOwner('6093ba0e2ab5775cfc01ed3e', '63639058685ba134c32bc495');
      expect(removeFromStructureLinkedSpyer).toBeCalledTimes(1);
      expect(deleteOfferSpyer).toBeCalledTimes(0);
    });

    it('should remove structure link with offer', async () => {
      const result = ({
        _id: '2345ba0e2ab5775cfc01ed4d',
        categories: {
          onlineProcedures: ['caf', 'cpam'],
          baseSkills: [],
          advancedSkills: [],
        },
        createdAt: 'Wed Mar 16 2022 14:29:11 GMT+0100 (heure normale d’Europe centrale)',
        updatedAt: 'Wed Mar 16 2022 17:29:11 GMT+0100 (heure normale d’Europe centrale)',
      } as PersonalOffer) as PersonalOfferDocument;
      jest.spyOn(usersService, 'getPersonalOfferInStructure').mockImplementationOnce(() => result);
      const removeFromStructureLinkedSpyer = jest.spyOn(usersService, 'removeFromStructureLinked');
      const deleteOfferSpyer = jest.spyOn(personalOffersService, 'delete');
      await controller.removeOwner('6093ba0e2ab5775cfc01ed3e', '63639058685ba134c32bc495');
      expect(removeFromStructureLinkedSpyer).toBeCalledTimes(1);
      expect(deleteOfferSpyer).toBeCalledTimes(1);
    });
  });

  it('should remove temp user', async () => {
    const res = controller.removeTempUser('6093ba0e2ab5775cfc01ed3e', 'tsfsf6296');
    expect(res).toBeTruthy();
  });

  it('should remove user from struct', async () => {
    const res = controller.removeOwner('6093ba0e2ab5775cfc01ed3e', 'tsfsf6296');
    expect(res).toBeTruthy();
  });

  it('should report any structure error', async () => {
    const res = controller.reportStructureError({ structureId: '6093ba0e2ab5775cfc01ed3e', content: null });
    expect(res).toBeTruthy();
  });
});
