import { HttpModule } from '@nestjs/axios';
import { HttpStatus } from '@nestjs/common';
import { getModelToken } from '@nestjs/mongoose';
import { Test, TestingModule } from '@nestjs/testing';
import { Types } from 'mongoose';
import { ConfigurationModule } from '../../configuration/configuration.module';
import { CreateJobDto } from '../dto/create-job.dto';
import { Job } from '../schemas/job.schema';
import { User } from '../schemas/user.schema';
import { JobsService } from '../services/jobs.service';
import { UsersService } from '../services/users.service';
import { JobsController } from './jobs.controller';

describe('JobsController', () => {
  let controller: JobsController;

  const jobServiceMock = {
    findAll: jest.fn(),
    findByName: jest.fn(),
    create: jest.fn(),
    findOne: jest.fn(),
    deleteOneId: jest.fn(),
    findAllUnvalidated: jest.fn(),
    createJobFromAdmin: jest.fn(),
    validate: jest.fn(),
    update: jest.fn(),
    mergeJob: jest.fn(),
    deleteInvalidJob: jest.fn(),
  };

  const userServiceMock = {
    populateJobswithUsers: jest.fn(),
    isJobLinkedtoUser: jest.fn(),
  };

  beforeEach(async () => {
    const module: TestingModule = await Test.createTestingModule({
      imports: [ConfigurationModule, HttpModule],
      providers: [
        {
          provide: JobsService,
          useValue: jobServiceMock,
        },
        {
          provide: UsersService,
          useValue: userServiceMock,
        },
        {
          provide: getModelToken('Job'),
          useValue: Job,
        },
        {
          provide: getModelToken('User'),
          useValue: User,
        },
      ],
      controllers: [JobsController],
    }).compile();

    controller = module.get<JobsController>(JobsController);
  });

  it('should be defined', () => {
    expect(controller).toBeDefined();
  });

  describe('findAll', () => {
    it('should findAll', async () => {
      jobServiceMock.findAll.mockResolvedValue([
        {
          _id: new Types.ObjectId('6231aefe76598527c8d0b5bc'),
          name: 'CNFS',
          validated: true,
          hasPersonalOffer: true,
          jobsGroup: null,
        },
        {
          _id: new Types.ObjectId('6231aefe76598527c8d0b5bc'),
          name: 'Formateur',
          validated: true,
          hasPersonalOffer: true,
          jobsGroup: null,
        },
      ]);
      const reply = await controller.findAll();
      expect(reply.length).toBe(2);
    });
  });

  describe('createJob', () => {
    it('should create job `Dev`', async () => {
      const newJob: CreateJobDto = {
        name: 'Dev',
        hasPersonalOffer: false,
        jobsGroup: null,
      };
      const newCreatedJob: Job = {
        name: 'Dev',
        validated: false,
        hasPersonalOffer: false,
        jobsGroup: null,
      };
      jobServiceMock.findByName.mockResolvedValueOnce(null);
      jobServiceMock.create.mockResolvedValueOnce(newCreatedJob);
      const req = { user: { _id: '6036721022462b001334c4bb' } };
      const createReply = await controller.createJob(req, newJob);
      expect(createReply).toEqual(newCreatedJob);
    });
    it('should throw error on already existing job `Dev`', async () => {
      const newJob: CreateJobDto = {
        name: 'Dev',
        hasPersonalOffer: false,
        jobsGroup: null,
      };
      jobServiceMock.findByName.mockResolvedValueOnce({
        _id: new Types.ObjectId('6231aefe76598527c8d0b5a7'),
        name: 'Metro',
        validated: false,
        hasPersonalOffer: false,
        jobsGroup: null,
      });
      try {
        const req = { user: { _id: '6036721022462b001334c4bb' } };
        await controller.createJob(req, newJob);
        expect(true).toBe(false);
      } catch (e) {
        expect(e.message).toBe('Job already exist');
        expect(e.status).toBe(422);
      }
    });
  });

  describe('Create Job', () => {
    it('should create a job', async () => {
      jobServiceMock.findByName.mockResolvedValueOnce(null);
      jobServiceMock.create.mockResolvedValueOnce({
        _id: new Types.ObjectId('6231aefe76598527c8d0b5ba'),
        name: 'Dev',
        validated: true,
      });
      const req = { user: { _id: '6036721022462b001334c4bb', role: 0 } };
      const reply = await controller.createJob(req, { name: 'Dev', hasPersonalOffer: true, jobsGroup: null });
      expect(reply).toBeTruthy();
    });
    it('should not create if job already exists', async () => {
      jobServiceMock.findByName.mockResolvedValueOnce({
        _id: new Types.ObjectId('6231aefe76598527c8d0b5ba'),
        name: 'Dev',
        validated: true,
      });
      const req = { user: { _id: '6036721022462b001334c4bb', role: 0 } };
      try {
        await controller.createJob(req, { name: 'Dev', hasPersonalOffer: true, jobsGroup: null });
      } catch (e) {
        expect(e.message).toBe('Job already exist');
        expect(e.status).toBe(HttpStatus.UNPROCESSABLE_ENTITY);
      }
    });
    it('should call create with send notification to true if admin', async () => {
      jobServiceMock.findByName.mockResolvedValueOnce(null);
      jobServiceMock.create.mockResolvedValueOnce({
        _id: new Types.ObjectId('6231aefe76598527c8d0b5ba'),
        name: 'Dev',
        validated: true,
      });
      const req = { user: { _id: '6036721022462b001334c4bb', role: 1 } };
      const job = { name: 'Dev', hasPersonalOffer: true, jobsGroup: null };
      const reply = await controller.createJob(req, job);
      expect(reply).toBeTruthy();
      expect(jobServiceMock.create).toHaveBeenCalledWith(job, true, true, false);
    });
  });
  describe('Validate Job', () => {
    it('should validate a given job', async () => {
      jobServiceMock.validate.mockResolvedValueOnce({
        _id: new Types.ObjectId('6231aefe76598527c8d0b5ba'),
        name: 'Dev',
        validated: true,
      });
      expect(await controller.validateJob({ id: '6231aefe76598527c8d0b5bc' })).toBeTruthy();
    });
  });
  describe('Get Jobs', () => {
    it('should call all validated jobs and populate them with users attached to it', async () => {
      const spyer = jest.spyOn(userServiceMock, 'populateJobswithUsers');
      jobServiceMock.findAll.mockResolvedValueOnce([
        {
          _id: new Types.ObjectId('6231aefe76598527c8d0b5bc'),
          name: 'Dev',
          validated: true,
        },
        {
          _id: new Types.ObjectId('6231aefe76598527c8d0b5bc'),
          name: 'Scrum',
          validated: true,
        },
      ]);
      await controller.findValidatedJobs();
      expect(jobServiceMock.findAll.mock.calls.length).toBe(2);
      expect(spyer.mock.calls.length).toBe(1);
    });

    it('should call all unvalidated jobs and populate them with users attached to it', async () => {
      const spyer = jest.spyOn(userServiceMock, 'populateJobswithUsers');
      jobServiceMock.findAllUnvalidated.mockResolvedValueOnce([
        {
          _id: new Types.ObjectId('6231aefe76598527c8d0b5bc'),
          name: 'Dev',
          validated: false,
        },
        {
          _id: new Types.ObjectId('6231aefe76598527c8d0b5bc'),
          name: 'Scrum',
          validated: false,
        },
      ]);
      await controller.findUnvalidatedJobs();
      expect(jobServiceMock.findAllUnvalidated.mock.calls.length).toBe(1);
      expect(spyer.mock.calls.length).toBe(2);
    });
  });
  describe('Edit Job', () => {
    it('should update job', async () => {
      jobServiceMock.update.mockResolvedValueOnce({
        _id: new Types.ObjectId('6231aefe76598527c8d0b5bc'),
        name: 'DevMod',
        validated: true,
      });
      expect(
        await controller.updateJob('6231aefe76598527c8d0b5bc', {
          name: 'DevMod',
          hasPersonalOffer: false,
          jobsGroup: null,
        })
      ).toBeTruthy();
    });

    it('should delete an unvalidated job and replace all its occurence with a chosen validated job', async () => {
      jobServiceMock.mergeJob.mockResolvedValueOnce({
        _id: new Types.ObjectId('6231aefe76598527c8d0b5ba'),
        name: 'Dev',
        validated: true,
      });
      jobServiceMock.deleteInvalidJob.mockResolvedValueOnce({
        n: 1,
        ok: 1,
        deletedCount: 1,
      });

      const reply = await controller.mergeJob({
        sourceJobId: '6231aefe76598527c8d0b5ba',
        targetJobId: '6231aefe76598527c8d0b5bc',
      });
      expect(reply).toBeTruthy();
    });
  });
  describe('Delete Job', () => {
    jobServiceMock.deleteOneId.mockResolvedValueOnce({
      n: 1,
      ok: 1,
      deletedCount: 1,
    });

    it('should delete job', async () => {
      jobServiceMock.findOne.mockResolvedValueOnce({
        _id: new Types.ObjectId('6231aefe76598527c8d0b5ba'),
        name: 'Dev',
        validated: true,
      });
      userServiceMock.isJobLinkedtoUser.mockResolvedValueOnce(false);
      const reply = await controller.deleteJob({ id: '6231aefe76598527c8d0b5ba' });
      expect(reply).toBeTruthy();
    });

    it('should not delete job if a user is linked', async () => {
      jobServiceMock.findOne.mockResolvedValueOnce({
        _id: new Types.ObjectId('6231aefe76598527c8d0b5bc'),
        name: 'Dev',
        validated: true,
      });
      userServiceMock.isJobLinkedtoUser.mockResolvedValueOnce(true);
      try {
        await controller.deleteJob({ id: '6231aefe76598527c8d0b5bc' });
        expect(true).toBe(false);
      } catch (e) {
        expect(e.message).toEqual('Cannot delete job. It has user(s) attached to it.');
        expect(e.status).toEqual(HttpStatus.FORBIDDEN);
      }
    });
  });
});
