import { HttpModule } from '@nestjs/axios';
import { HttpStatus } from '@nestjs/common';
import { getModelToken } from '@nestjs/mongoose';
import { Test, TestingModule } from '@nestjs/testing';
import { AxiosResponse } from 'axios';
import { Types } from 'mongoose';
import { ConfigurationModule } from '../../configuration/configuration.module';
import { MailerService } from '../../mailer/mailer.service';
import { CreateEmployerDto } from '../dto/create-employer.dto';
import { EmployerDocument } from '../schemas/employer.schema';
import { User } from '../schemas/user.schema';
import { EmployerSearchService } from './employer-search.service';
import { EmployerService } from './employer.service';
import { UsersService } from './users.service';

describe('EmployerService', () => {
  let service: EmployerService;
  let mailer: MailerService;

  const mockEmployerSearchService = {
    indexEmployer: jest.fn(),
    search: jest.fn(),
    dropIndex: jest.fn(),
    createEmployerIndex: jest.fn(),
    deleteIndex: jest.fn(),
  };
  const mockEmployerModel = {
    create: jest.fn(),
    findOne: jest.fn(),
    findById: jest.fn(),
    deleteOne: jest.fn(),
    find: jest.fn(() => mockEmployerModel),
    sort: jest.fn(() => mockEmployerModel),
    exec: jest.fn(),
  };
  const mockUserService = {
    replaceEmployers: jest.fn(),
    getAdmins: jest.fn(),
  };
  beforeEach(async () => {
    const module: TestingModule = await Test.createTestingModule({
      imports: [ConfigurationModule, HttpModule],
      providers: [
        EmployerService,
        {
          provide: EmployerSearchService,
          useValue: mockEmployerSearchService,
        },
        {
          provide: getModelToken('Employer'),
          useValue: mockEmployerModel,
        },
        {
          provide: UsersService,
          useValue: mockUserService,
        },
        {
          provide: getModelToken('User'),
          useValue: User,
        },
        MailerService,
      ],
    }).compile();
    service = module.get<EmployerService>(EmployerService);
    mailer = module.get<MailerService>(MailerService);
  });

  it('should be defined', () => {
    expect(service).toBeDefined();
  });

  it('findAll', async () => {
    mockEmployerModel.sort.mockReturnThis();
    mockEmployerModel.exec.mockResolvedValueOnce([
      {
        _id: new Types.ObjectId('6231aefe76598527c8d0b5bc'),
        name: 'Metro',
        validated: true,
      },
      {
        _id: new Types.ObjectId('6231aefe76598527c8d0b5bc'),
        name: 'Sopra',
        validated: true,
      },
    ]);
    const reply = await service.findAll();
    expect(reply.length).toBe(2);
  });
  it('findOne', async () => {
    mockEmployerModel.findById.mockReturnThis();
    mockEmployerModel.exec.mockResolvedValueOnce({
      _id: new Types.ObjectId('6231aefe76598527c8d0b5bc'),
      name: 'Sopra',
      validated: true,
    });
    const reply = await service.findOne('6231aefe76598527c8d0b5bc');
    expect(reply).toBeTruthy();
  });
  it('findAllValidated', async () => {
    mockEmployerModel.find.mockReturnThis();
    mockEmployerModel.exec.mockResolvedValueOnce([
      {
        _id: new Types.ObjectId('6231aefe76598527c8d0b5bc'),
        name: 'Sopra',
        validated: true,
      },
    ]);
    const reply = await service.findAllValidated();
    expect(reply.length).toBe(1);
  });

  it('findAllUnvalidated', async () => {
    mockEmployerModel.find.mockReturnThis();
    mockEmployerModel.exec.mockResolvedValueOnce([
      {
        _id: new Types.ObjectId('6231aefe76598527c8d0b5bc'),
        name: 'Sopra',
        validated: false,
      },
    ]);
    const reply = await service.findAllValidated();
    expect(reply.length).toBe(1);
  });

  it('finds all unvalidated employers', async () => {
    mockEmployerModel.find.mockReturnThis();
    mockEmployerModel.exec.mockResolvedValueOnce([
      {
        _id: new Types.ObjectId('6231aefe76598527c8d0b5bc'),
        name: 'Metro',
        validated: false,
      },
    ]);
    const reply = await service.findAllUnvalidated();
    expect(reply[0].validated).toBeFalsy;
  });

  it('finds all validated employers', async () => {
    mockEmployerModel.find.mockReturnThis();
    mockEmployerModel.exec.mockResolvedValueOnce([
      {
        _id: new Types.ObjectId('6231aefe76598527c8d0b5bc'),
        name: 'Metro',
        validated: true,
      },
      {
        _id: new Types.ObjectId('6231aefe76598527c8d0b5b2'),
        name: 'Sopra',
        validated: true,
      },
    ]);
    const reply = await service.findAllUnvalidated();
    expect(reply.length).toBe(2);
  });

  it('findByName', async () => {
    mockEmployerModel.findOne.mockReturnThis();
    mockEmployerModel.exec.mockResolvedValueOnce({
      _id: new Types.ObjectId('6231aefe76598527c8d0b5bc'),
      name: 'Sopra',
      validated: true,
    });
    const reply = await service.findByName('Sopra');
    expect(reply).toBeTruthy();
  });

  describe('createEmployer', () => {
    it('create', async () => {
      mockEmployerModel.create.mockResolvedValue({
        _id: new Types.ObjectId('6231aefe76598527c8d0b5bc'),
        name: 'Sopra',
        validated: false,
      });
      mockEmployerModel.findOne.mockReturnThis();
      mockEmployerModel.exec.mockResolvedValueOnce({
        _id: new Types.ObjectId('6231aefe76598527c8d0b5bc'),
        name: 'Sopra',
        validated: false,
      });
      const createEmployer: CreateEmployerDto = {
        name: 'Sopra',
      };
      jest.spyOn(service, 'sendAdminCreateNotification').mockResolvedValueOnce();
      const reply = await service.create(createEmployer);
      expect(reply).toBeTruthy();
      expect(service.sendAdminCreateNotification).toBeCalledTimes(1);
    });

    it('should create validated employer and not send email to admins', async () => {
      mockEmployerModel.create.mockResolvedValue({
        _id: new Types.ObjectId('6231aefe76598527c8d0b5bc'),
        name: 'Sopra',
        validated: false,
      });
      mockEmployerModel.findOne.mockReturnThis();
      mockEmployerModel.exec.mockResolvedValueOnce({
        _id: new Types.ObjectId('6231aefe76598527c8d0b5bc'),
        name: 'Sopra',
        validated: false,
      });
      const createEmployer: CreateEmployerDto = {
        name: 'Sopra',
      };
      const reply = await service.create(createEmployer, true, false);
      expect(reply).toBeTruthy();
    });
  });

  it('delete', async () => {
    mockEmployerModel.findOne.mockReturnThis();
    mockEmployerModel.exec.mockResolvedValueOnce({
      _id: new Types.ObjectId('6231aefe76598527c8d0b5bc'),
      name: 'Sopra',
      validated: true,
    });
    mockEmployerModel.deleteOne.mockReturnThis();
    mockEmployerModel.exec.mockResolvedValueOnce({
      n: 1,
      ok: 1,
      deletedCount: 1,
    });
    mockEmployerSearchService.deleteIndex.mockResolvedValueOnce({});
    const reply = await service.deleteByName('Sopra');
    expect(reply).toBeTruthy();
  });

  describe('Search', () => {
    it('searchByName', async () => {
      mockEmployerSearchService.search.mockResolvedValue({ name: 'Sopra' });
      const reply = await service.searchByName('Sopra');
      expect(reply).toBeTruthy();
    });

    it('initEmployerIndex', async () => {
      mockEmployerSearchService.dropIndex.mockResolvedValueOnce({});
      mockEmployerSearchService.createEmployerIndex.mockResolvedValueOnce({});
      mockEmployerModel.sort.mockReturnThis();
      mockEmployerModel.exec.mockResolvedValueOnce([
        {
          _id: new Types.ObjectId('6231aefe76598527c8d0b5bc'),
          name: 'Metro',
          validated: true,
        },
        {
          _id: new Types.ObjectId('6231aefe76598527c8d0b5bc'),
          name: 'Sopra',
          validated: true,
        },
      ]);
      mockEmployerSearchService.indexEmployer
        .mockResolvedValueOnce({
          _id: new Types.ObjectId('6231aefe76598527c8d0b5bc'),
          name: 'Metro',
          validated: true,
        } as EmployerDocument)
        .mockResolvedValueOnce({
          _id: new Types.ObjectId('6231aefe76598527c8d0b5bc'),
          name: 'Sopra',
          validated: true,
        } as EmployerDocument);
      const reply = await service.initEmployerIndex();
      expect(reply).toBeTruthy();
      expect(reply.length).toBe(2);
    });
  });

  describe('mergeEmployer', () => {
    it('should delete source employer', async () => {
      const result = {
        _id: new Types.ObjectId('623aed68c5d45b6fbbaa7e61'),
        name: 'Metro',
        validated: true,
      };
      mockEmployerModel.findById.mockReturnThis();
      mockEmployerModel.exec
        .mockResolvedValueOnce({
          _id: new Types.ObjectId('623aed68c5d45b6fbbaa7e60'),
          name: 'Sopra',
          validated: false,
        })
        .mockResolvedValueOnce(result)
        .mockResolvedValueOnce({
          _id: new Types.ObjectId('623aed68c5d45b6fbbaa7e60'),
          name: 'Sopra',
          validated: false,
        });
      mockUserService.replaceEmployers.mockResolvedValueOnce(null);
      const reply = await service.mergeEmployer({
        sourceEmployerId: '623aed68c5d45b6fbbaa7e60',
        targetEmployerId: '623aed68c5d45b6fbbaa7e61',
      });
      expect(reply).toEqual(result);
    });
    it('should return error if delete source employer issue', async () => {
      const result = {
        _id: new Types.ObjectId('623aed68c5d45b6fbbaa7e61'),
        name: 'Metro',
        validated: true,
      };
      mockEmployerModel.findById.mockReturnThis();
      mockEmployerModel.exec.mockResolvedValueOnce(null).mockResolvedValueOnce(result);
      jest.spyOn(service, 'deleteInvalidEmployer').mockRejectedValueOnce({});
      mockUserService.replaceEmployers.mockResolvedValueOnce(null);
      try {
        await service.mergeEmployer({
          sourceEmployerId: '623aed68c5d45b6fbbaa7e60',
          targetEmployerId: '623aed68c5d45b6fbbaa7e61',
        });
      } catch (e) {
        expect(e.message).toBe('Cannot operate on employer.');
        expect(e.status).toBe(HttpStatus.NOT_FOUND);
      }
    });
  });

  describe('sendAdminCreateNotification', () => {
    it('should sendAdminCreateNotification', async () => {
      const employer = {
        _id: new Types.ObjectId('6231aefe76598527c8d0b5bc'),
        name: 'Metro',
        validated: true,
      };
      const result: AxiosResponse = {
        data: {
          status: 200,
          content: {
            success: true,
            response: [7010],
          },
        },
        status: 200,
        statusText: 'OK',
        headers: {},
        config: {},
      };
      jest.spyOn(mailer, 'send').mockResolvedValueOnce(result);
      const spyer = jest.spyOn(mailer, 'send');
      mockUserService.getAdmins.mockResolvedValueOnce([
        {
          _id: '6231aefe76598527c8d0b5bc',
          validationToken:
            'cf1c74c22cedb6b575945098db42d2f493fb759c9142c6aff7980f252886f36ee086574ee99a06bc99119079257116c959c8ec870949cebdef2b293666dbca42',
          emailVerified: true,
          email: 'admin@admin.com',
          password: '$2a$12$vLQjJ9zAWyUwiFLeQDa6w.XzrlgPBhw.2GWrjog/yuEjIaZnQwmZu',
          role: 0,
          name: 'admin',
          surname: 'admin',
          personalOffers: [],
        },
      ]);
      await service.sendAdminCreateNotification(
        employer as EmployerDocument,
        'adminEmployerCreate.ejs',
        'adminEmployerCreate.json'
      );
      expect(spyer.mock.calls.length).toBe(1);
    });
  });

  describe('validate', () => {
    it('should validate employer', async () => {
      mockEmployerModel.findById.mockReturnThis();
      mockEmployerModel.exec.mockResolvedValueOnce({
        _id: new Types.ObjectId('623aed68c5d45b6fbbaa7e60'),
        name: 'Sopra',
        validated: false,
        save: jest.fn().mockResolvedValueOnce(null),
      });
      const employer = await service.validate('623aed68c5d45b6fbbaa7e60');
      expect(employer.validated).toBe(true);
    });
    it('should throw exception', async () => {
      mockEmployerModel.findById.mockReturnThis();
      mockEmployerModel.exec.mockResolvedValueOnce(null);
      try {
        await service.validate('623aed68c5d45b6fbbaa7e60');
        expect(true).toBe(false);
      } catch (e) {
        expect(e.message).toBe('Cannot validate employer. It might have been already validate');
        expect(e.status).toBe(HttpStatus.NOT_FOUND);
      }
    });
  });

  describe('update', () => {
    it('should update employer', async () => {
      mockEmployerModel.findById.mockReturnThis();
      mockEmployerModel.exec.mockResolvedValueOnce({
        _id: new Types.ObjectId('623aed68c5d45b6fbbaa7e60'),
        name: 'Sopraaa',
        validated: false,
        save: jest.fn().mockResolvedValueOnce(null),
      });
      const employer = await service.update('623aed68c5d45b6fbbaa7e60', { name: 'Sopra' });
      expect(employer.name).toBe('Sopra');
    });
    it('should throw exception', async () => {
      mockEmployerModel.findById.mockReturnThis();
      mockEmployerModel.exec.mockResolvedValueOnce(null);
      try {
        await service.update('623aed68c5d45b6fbbaa7e60', { name: 'Sopra' });
        expect(true).toBe(false);
      } catch (e) {
        expect(e.message).toBe('Cannot edit employer. It was not found in database.');
        expect(e.status).toBe(HttpStatus.NOT_FOUND);
      }
    });
  });

  describe('deleteInvalidEmployer', () => {
    it('should delete invalid employer', async () => {
      mockEmployerModel.findById.mockReturnThis();
      mockEmployerModel.exec.mockResolvedValueOnce({
        _id: new Types.ObjectId('623aed68c5d45b6fbbaa7e60'),
        name: 'Sopra',
        validated: false,
        save: jest.fn().mockResolvedValueOnce(null),
      });
      mockEmployerSearchService.deleteIndex.mockResolvedValueOnce(null);
      mockEmployerModel.deleteOne.mockReturnThis();
      mockEmployerModel.exec.mockResolvedValueOnce({
        _id: new Types.ObjectId('623aed68c5d45b6fbbaa7e60'),
        name: 'Sopra',
        validated: false,
        save: jest.fn().mockResolvedValueOnce(null),
      });
      await service.deleteInvalidEmployer('6231aefe76598527c8d0b5bc');
      expect(mockEmployerSearchService.deleteIndex).toBeCalled();
      expect(mockEmployerModel.deleteOne).toBeCalled();
    });
  });

  describe('deleteOneId', () => {
    it('should delete ', async () => {
      mockEmployerModel.findOne.mockReturnThis();
      mockEmployerModel.exec.mockResolvedValueOnce({
        _id: '6231aefe76598527c8d0b5bc',
        validationToken:
          'cf1c74c22cedb6b575945098db42d2f493fb759c9142c6aff7980f252886f36ee086574ee99a06bc99119079257116c959c8ec870949cebdef2b293666dbca42',
        emailVerified: true,
        email: 'admin@admin.com',
        password: '$2a$12$vLQjJ9zAWyUwiFLeQDa6w.XzrlgPBhw.2GWrjog/yuEjIaZnQwmZu',
        role: 0,
        name: 'admin',
        surname: 'admin',
        personalOffers: [],
        deleteOne: jest.fn().mockResolvedValueOnce({
          _id: '6231aefe76598527c8d0b5bc',
          validationToken:
            'cf1c74c22cedb6b575945098db42d2f493fb759c9142c6aff7980f252886f36ee086574ee99a06bc99119079257116c959c8ec870949cebdef2b293666dbca42',
          emailVerified: true,
          email: 'admin@admin.com',
          password: '$2a$12$vLQjJ9zAWyUwiFLeQDa6w.XzrlgPBhw.2GWrjog/yuEjIaZnQwmZu',
          role: 0,
          name: 'admin',
          surname: 'admin',
          personalOffers: [],
        }),
      });
      expect(await service.deleteOneId('6231aefe76598527c8d0b5bc')).toStrictEqual({
        _id: '6231aefe76598527c8d0b5bc',
        validationToken:
          'cf1c74c22cedb6b575945098db42d2f493fb759c9142c6aff7980f252886f36ee086574ee99a06bc99119079257116c959c8ec870949cebdef2b293666dbca42',
        emailVerified: true,
        email: 'admin@admin.com',
        password: '$2a$12$vLQjJ9zAWyUwiFLeQDa6w.XzrlgPBhw.2GWrjog/yuEjIaZnQwmZu',
        role: 0,
        name: 'admin',
        surname: 'admin',
        personalOffers: [],
      });
    });
    it('should throw an error ', async () => {
      mockEmployerModel.findOne.mockReturnThis();
      mockEmployerModel.exec.mockResolvedValueOnce(null);

      try {
        expect(await service.deleteOneId('6231aefe76598527c8d0b5bc')).toStrictEqual({
          _id: '6231aefe76598527c8d0b5bc',
          validationToken:
            'cf1c74c22cedb6b575945098db42d2f493fb759c9142c6aff7980f252886f36ee086574ee99a06bc99119079257116c959c8ec870949cebdef2b293666dbca42',
          emailVerified: true,
          email: 'admin@admin.com',
          password: '$2a$12$vLQjJ9zAWyUwiFLeQDa6w.XzrlgPBhw.2GWrjog/yuEjIaZnQwmZu',
          role: 0,
          name: 'admin',
          surname: 'admin',
          personalOffers: [],
        });
        expect(true).toBe(false);
      } catch (e) {
        expect(e.message).toBe('Invalid employer id');
        expect(e.status).toBe(HttpStatus.BAD_REQUEST);
      }
    });
  });
});
