import { HttpModule } from '@nestjs/axios';
import { getModelToken } from '@nestjs/mongoose';
import { Test, TestingModule } from '@nestjs/testing';
import { Types } from 'mongoose';
import { multipleUsers } from '../../../test/mock/data/users.mock.data';
import { ConfigurationModule } from '../../configuration/configuration.module';
import { IUserRegistry } from '../interfaces/userRegistry.interface';
import { EmployerService } from './employer.service';
import { JobsGroupsService } from './jobsGroups.service';
import { UserRegistrySearchService } from './userRegistry-search.service';
import { UserRegistryService } from './userRegistry.service';

describe('userRegistryService', () => {
  let service: UserRegistryService;
  const mockUserRegistryModel = {
    find: jest.fn(() => mockUserRegistryModel),
    where: jest.fn(() => mockUserRegistryModel),
    equals: jest.fn(() => mockUserRegistryModel),
    populate: jest.fn(() => mockUserRegistryModel),
    collation: jest.fn(() => mockUserRegistryModel),
    sort: jest.fn(() => mockUserRegistryModel),
    select: jest.fn(() => mockUserRegistryModel),
    exec: jest.fn(() => mockUserRegistryModel),
    count: jest.fn(() => mockUserRegistryModel),
    limit: jest.fn(() => mockUserRegistryModel),
  };

  const mockUserRegistrySearchService = {
    search: jest.fn(),
    dropIndex: jest.fn(),
    deleteIndex: jest.fn(),
    createUserRegistryIndex: jest.fn(),
    indexUserRegistry: jest.fn(),
  };
  const mockEmployersService = {
    findByName: jest.fn(),
  };
  const mockJobsGroupsService = {
    findByName: jest.fn(),
  };

  beforeEach(async () => {
    const module: TestingModule = await Test.createTestingModule({
      imports: [ConfigurationModule, HttpModule],
      providers: [
        UserRegistryService,
        {
          provide: UserRegistrySearchService,
          useValue: mockUserRegistrySearchService,
        },
        {
          provide: getModelToken('User'),
          useValue: mockUserRegistryModel,
        },
        {
          provide: JobsGroupsService,
          useValue: mockJobsGroupsService,
        },
        {
          provide: EmployerService,
          useValue: mockEmployersService,
        },
      ],
    }).compile();
    service = module.get<UserRegistryService>(UserRegistryService);
  });

  it('should be defined', () => {
    expect(service).toBeDefined();
  });

  describe('findAll', () => {
    const result: IUserRegistry[] = [
      {
        _id: new Types.ObjectId('6319dfa79672971e1f8fe1b7'),
        surname: 'ADMIN',
        name: 'Admin',
        employer: {
          name: 'Pimms',
          validated: true,
        },
        job: {
          hasPersonalOffer: true,
          name: 'CNFS',
          validated: true,
        },
      },
    ] as IUserRegistry[];
    it('should findAll UserRegistry for indexation', async () => {
      mockUserRegistryModel.exec.mockResolvedValueOnce(result);
      expect(await service.findAllForIndexation()).toBe(result);
    });

    it('should findAll UserRegistry count', async () => {
      mockUserRegistryModel.exec.mockResolvedValueOnce(result.length);
      expect(await service.countAllUserRegistry()).toBe(result.length);
    });
    it('should findAllUserRegistry with page number 1', async () => {
      const res = { count: 1, docs: result };
      mockUserRegistryModel.exec.mockResolvedValueOnce(1);
      mockUserRegistryModel.exec.mockResolvedValueOnce(result);
      expect(await service.findAllUserRegistry(1)).toStrictEqual(res);
    });
  });
  describe('find with filter', () => {
    const result: IUserRegistry[] = [
      {
        _id: new Types.ObjectId('6319dfa79672971e1f8fe1b7'),
        surname: 'ADMIN',
        name: 'Admin',
        employer: {
          name: 'Pimms',
          validated: true,
        },
        job: {
          hasPersonalOffer: true,

          name: 'CNFS',
          validated: true,
        },
      },
    ] as IUserRegistry[];
    const res = {
      count: 1,
      docs: [
        {
          _id: new Types.ObjectId('6319dfa79672971e1f8fe1b7'),
          surname: 'ADMIN',
          name: 'Admin',
          employer: {
            name: 'Pimms',
            validated: true,
          },
          job: {
            hasPersonalOffer: true,
            name: 'CNFS',
            validated: true,
          },
        },
      ],
    };
    it('should findUsersByNameEmployerOrJobsGroup with string param and get a result', async () => {
      mockUserRegistryModel.exec.mockResolvedValueOnce(result);
      mockUserRegistrySearchService.search.mockResolvedValueOnce(result);
      expect(await service.findUsersByNameEmployerOrJobsGroup('adm', 1)).toStrictEqual(res);
    });
    it('should findUsersByNameEmployerOrJobsGroup with string param and get no result', async () => {
      const emptyRes = { count: 0, docs: [] };
      mockUserRegistryModel.exec.mockResolvedValueOnce([]);
      mockUserRegistrySearchService.search.mockResolvedValueOnce([]);
      expect(await service.findUsersByNameEmployerOrJobsGroup('azerty', 1)).toStrictEqual(emptyRes);
    });
    it('should findUsersByNameEmployerOrJobsGroup with no string param and filters', async () => {
      const res = { count: 1, docs: [multipleUsers[1]] };
      const jobGroupList = [
        {
          _id: new Types.ObjectId('6442ad5a5b3fd128a5c0e3b4'),
          name: 'Technique',
        },
      ];
      const employerList = [{ name: 'Métropole', validated: true }];
      mockUserRegistrySearchService.search.mockResolvedValueOnce(multipleUsers);
      mockJobsGroupsService.findByName.mockResolvedValueOnce(jobGroupList[0]);
      mockEmployersService.findByName.mockResolvedValueOnce(employerList[0]);
      mockUserRegistryModel.exec.mockResolvedValueOnce(multipleUsers);
      expect(await service.findUsersByNameEmployerOrJobsGroup('', 1, ['Technique'], ['Métropole'])).toStrictEqual(res);
    });

    it('should findUsersByNameEmployerOrJobsGroup with string param and filters', async () => {
      const res = { count: 1, docs: [multipleUsers[1]] };
      const jobGroupList = [
        {
          _id: new Types.ObjectId('6442ad5a5b3fd128a5c0e3b4'),
          name: 'Technique',
        },
      ];
      const employerList = [{ name: 'Métropole', validated: true }];
      mockUserRegistrySearchService.search.mockResolvedValueOnce([multipleUsers[1]]);
      mockJobsGroupsService.findByName.mockResolvedValueOnce(jobGroupList[0]);
      mockEmployersService.findByName.mockResolvedValueOnce(employerList[0]);
      mockUserRegistryModel.exec.mockResolvedValueOnce(multipleUsers);
      expect(await service.findUsersByNameEmployerOrJobsGroup('Guil', 1, ['Technique'], ['Métropole'])).toStrictEqual(
        res
      );
    });
    it('should findUsersByNameEmployerOrJobsGroup with string param and filters and return empty', async () => {
      const res = { count: 0, docs: [] };
      const jobsGroupList = [
        {
          _id: new Types.ObjectId('6442ad5a5b3fd128a5c0e3b4'),
          name: 'Technique',
        },
      ];
      const employerList = [{ name: 'Métropole', validated: true }];
      mockUserRegistrySearchService.search.mockResolvedValueOnce([]);
      mockJobsGroupsService.findByName.mockResolvedValueOnce(jobsGroupList[0]);
      mockEmployersService.findByName.mockResolvedValueOnce(employerList[0]);
      mockUserRegistryModel.exec.mockResolvedValueOnce([]);
      expect(
        await service.findUsersByNameEmployerOrJobsGroup('azerrttt', 1, ['Technique'], ['Métropole'])
      ).toStrictEqual(res);
    });
    it('should findUsersByNameEmployerOrJobsGroup with string param and one employer filter', async () => {
      const res = { count: 2, docs: [multipleUsers[0], multipleUsers[2]] };
      const employerList = [{ name: 'CAF', validated: true }];
      mockUserRegistrySearchService.search.mockResolvedValueOnce(multipleUsers);
      mockEmployersService.findByName.mockResolvedValueOnce(employerList[0]);
      mockUserRegistryModel.exec.mockResolvedValueOnce(multipleUsers);
      expect(await service.findUsersByNameEmployerOrJobsGroup('a', 1, [], ['CAF'])).toStrictEqual(res);
    });

    it('should findUsersByNameEmployerOrJobsGroup with no string param and one jobsGroup filter', async () => {
      const res = { count: 1, docs: [multipleUsers[1]] };
      const jobGroupList = [
        {
          _id: new Types.ObjectId('6442ad5a5b3fd128a5c0e3b4'),
          name: 'Technique',
        },
      ];
      mockUserRegistrySearchService.search.mockResolvedValueOnce(multipleUsers);
      mockJobsGroupsService.findByName.mockResolvedValueOnce(jobGroupList[0]);
      mockUserRegistryModel.exec.mockResolvedValueOnce(multipleUsers);
      expect(await service.findUsersByNameEmployerOrJobsGroup('', 1, ['Technique'], [])).toStrictEqual(res);
    });
  });

  describe('Search', () => {
    it('should initUserRegistryIndex', async () => {
      mockUserRegistryModel.exec.mockResolvedValueOnce(multipleUsers);
      mockUserRegistrySearchService.dropIndex.mockResolvedValueOnce({});
      mockUserRegistrySearchService.createUserRegistryIndex.mockResolvedValueOnce({});
      mockUserRegistrySearchService.indexUserRegistry.mockResolvedValueOnce({});
      expect(await service.initUserRegistryIndex()).toBe(multipleUsers);
    });

    it('should findAllUserRegistry with page number 1', async () => {
      mockUserRegistrySearchService.search.mockResolvedValueOnce([multipleUsers[0]]);

      expect(await service.searchByNameAndSurname('adm')).toStrictEqual([multipleUsers[0]]);
    });
  });
});
